using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Utils.Log.Event;
using System.ComponentModel;
using System.IO;
using Utils.Log.Exceptions;
using System.Timers;

namespace Utils.Log
{
    /// <summary>
    /// Conatain a list of log entry
    /// </summary>
    public class LogSet : IEnumerator, IDisposable
    {
        #region Variables
        /// <summary>
        /// Timer to check whether the file changed
        /// </summary>
        private Timer vTimer = new Timer(5000);
        /// <summary>
        /// List of log entries
        /// </summary>
        private List<LogRow> vList = new List<LogRow>();
        /// <summary>
        /// Path if the log file
        /// </summary>
        private string vFilePath = "";
        /// <summary>
        /// IEnumarator index
        /// </summary>
        private int vIndex = -1;
        /// <summary>
        /// Indicates whether the object is analyzing the log file
        /// </summary>
        private bool vGuarding = false;
        /// <summary>
        /// Indicates the last modification time of the file
        /// </summary>
        private DateTime vLastFileModification;
        /// <summary>
        /// Indicates the size of the file
        /// </summary>
        private float vFileSize = 0;
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public LogSet() { }
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="_FilePath">Indicates the path of the Log file</param>
        public LogSet(string _FilePath) { this.FilePath = _FilePath; }
        /// <summary>
        /// Dispose the ressouces
        /// </summary>
        ~LogSet() { this.Dispose(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the list of log entries
        /// </summary>
        public List<LogRow> List
        {
            get { return vList; }
            set { vList = value; }
        }
        /// <summary>
        /// Gets the elements number _index from the log entry list
        /// </summary>
        /// <param name="_index">Number of the log entry</param>
        /// <returns>The log entry</returns>
        /// <exception cref="IndexOutOfRangeException">If the index if too big or too small</exception>
        public LogRow this[int _index]
        {
            get
            {
                if (_index > vList.Count || _index < 0)
                    throw new IndexOutOfRangeException("The index does not exist in the log entry list");
                else
                return vList[_index];
                    
            }
        }
        /// <summary>
        /// Gets the number of log entries in the list
        /// </summary>
        public int Count { get { return vList.Count; } }
        /// <summary>
        /// Gets or sets the path of the log file
        /// </summary>
        public string FilePath
        {
            get { return vFilePath; }
            set { vFilePath = value; }
        }
        /// <summary>
        /// Gets the last modification of the log file
        /// </summary>
        public DateTime LastFileModification
        {
            get { return vLastFileModification; }
            private set { vLastFileModification = value; }
        }
        /// <summary>
        /// Gets the dates contained in the logs
        /// </summary>
        public DateTime[] ListOfDates
        {
            get
            {
                List<DateTime> Dates = new List<DateTime>();
                foreach (LogRow lr in this.List)
                {
                    DateTime lrDate = new DateTime(lr.Date.Year, lr.Date.Month, lr.Date.Day);
                    if (!Dates.Contains(lrDate))
                        Dates.Add(lrDate);
                }
                DateTime[] dt = new DateTime[Dates.Count];
                for (int i = 0; i < dt.Length; i++)
                {
                    dt[i] = Dates[i];
                }
                return dt;
            }
        }
        /// <summary>
        /// Gets the log file size in Ko
        /// </summary>
        public float FileSize { get { return this.vFileSize / 1024; } }
        #endregion

        #region Methods
        /// <summary>
        /// Fill the list of log entries from the file path
        /// </summary>
        /// <exception cref="LogException">The file does not contain expected log entries</exception>
        public void Fill()
        {
            LogRow lr = new LogRow();
            string vBuffer = ReadFile();
            vBuffer = vBuffer.Replace(Log.WLogTag, "");
            string[] opt = { Log.LineSeparator };
            string[] vEntryList = vBuffer.Split(opt, StringSplitOptions.RemoveEmptyEntries);

            foreach (string vLine in vEntryList)
            {
                try
                {
                    lr = LogRow.Parse(vLine);
                    this.Add(lr);
                }
                catch (BadLogParseException ex)
                { throw new LogException(ex.Message); }
            }
        }
        /// <summary>
        /// Add a log entry in the list
        /// </summary>
        /// <param name="_LogRow">Log entry to add in the list</param>       
        public void Add(LogRow _LogRow) { this.vList.Add(_LogRow); }
        /// <summary>
        /// Add a log entry list in the list
        /// </summary>
        /// <param name="_LogSet">Log entry list to add in the list</param>
        public void AddRange(LogSet _LogSet) { this.vList.AddRange(_LogSet.vList); }
        /// <summary>
        /// Delete the log entry at the index _index
        /// </summary>
        /// <param name="_index">_index of the log entry to delete</param>        
        /// <exception cref="ArgumentOutOfRangeException">Impossible to remove the entry at the given position</exception>
        public void RemoveAt(int _index)
        {
            try { this.List.RemoveAt(_index); }
            catch (ArgumentOutOfRangeException)
            {
                throw new ArgumentOutOfRangeException("Impossible to remove the entry at position " + _index);
            }
        }
        /// <summary>
        /// Save the list into a log file
        /// </summary>
        /// <exception cref="IOException">The exception that is thrown when an I/O error occurs.</exception>
        /// <exception cref="Exception">An unexpected error occured</exception>
        public void Save()
        {
            string vToWrite = Log.WLogTag + Log.LineSeparator;

            foreach (LogRow lr in this.List)
                vToWrite += lr;
            try
            {
                FileInfo file = new FileInfo(this.FilePath);
                FileStream fs = file.Create();                
                StreamWriter sw = new StreamWriter(fs);                
                sw.Write(vToWrite);                
                sw.Close();
                fs.Close();
            }
            catch (IOException) { throw; }
            catch (Exception) { throw; }
        }
        /// <summary>
        /// Reset the whole list of log entries
        /// </summary>
        public void Clear()
        {
            vList.Clear();
        }
        /// <summary>
        /// Search all the log of this date
        /// </summary>
        /// <param name="_Time">Date of the logs</param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(DateTime _Time)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogRow lr in this.List)
            {
                if (lr.Date.Year == _Time.Year
                    && lr.Date.Month == _Time.Month
                    && lr.Date.Day == _Time.Day)
                    _Result.Add(lr);
            }
            return _Result;
        }
        /// <summary>
        /// Search all the log of this type
        /// </summary>
        /// <param name="_Type">Type of log </param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(LogType _Type)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogRow lr in this.List)
            {
                if (lr.Type == _Type)
                    _Result.Add(lr);
            }
            return _Result;
        }
        /// <summary>
        /// Search all the log of these types
        /// </summary>
        /// <param name="_Types">Array of log types</param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(List<LogType> _Types)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogType t in _Types)
            {
                _Result.AddRange(this.Search(t));
            }
            return _Result;
        }
        /// <summary>
        /// Analyze the log file and throws the event FileChanged if the file changed
        /// </summary>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        public void StartGuard()
        {
            try{
            if (!vGuarding)
            {
                SetupTimer();
                vTimer.Start();
                vGuarding = true;
            }
        }
        catch (ArgumentException) { throw; }
        catch (NotSupportedException) { throw; }
        catch (PathTooLongException) { throw; }
        catch (UnauthorizedAccessException) { throw; }
        catch (Exception) { throw; }
        }
        /// <summary>
        /// Stop the analyze the log file.
        /// </summary>
        public void StopGuard()
        {
            if (vGuarding)
            {
                vTimer.Stop();
                vGuarding = false;
            }
        }
        /// <summary>
        /// Give the next element of the list or throw an exception if does not exist
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">If the next element does not exist</exception>
        public LogRow Next()
        {
            try { return this.List[++vIndex]; }
            catch (ArgumentOutOfRangeException) { throw; }
        }
        /// <summary>
        /// Give the previous element of the list or throw an exception if does not exist
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">If the previous element does not exist</exception>
        public LogRow Previous()
        {
            try { return this.List[--vIndex]; }
            catch (ArgumentOutOfRangeException) { throw; }
        }
        /// <summary>
        /// Give the next element of the list or NULL if does not exist
        /// </summary>
        public LogRow TryNext()
        {
            try { return this.Next(); }
            catch (ArgumentOutOfRangeException) { return null; }
        }
        /// <summary>
        /// Give the previous element of the list or NULL if does not exist
        /// </summary>
        public LogRow TryPrevious()
        {
            try { return this.Previous(); }
            catch (ArgumentOutOfRangeException) { return null; }
        }
        /// <summary>
        /// Read the file and returns the text
        /// </summary>
        /// <returns>The text of the file</returns>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        private string ReadFile()
        {
            try
            {
                FileInfo fi = new FileInfo(this.FilePath);
                this.vFileSize = fi.Length;
                if (!fi.Exists)
                    throw new FileNotFoundException("The log file does not exists at " + this.FilePath);
                StreamReader sr = fi.OpenText();
                string ret = sr.ReadToEnd();
                sr.Close();
                return ret;
            }
            catch (ArgumentException) { throw; }
            catch (NotSupportedException) { throw; }
            catch (PathTooLongException) { throw; }
            catch (UnauthorizedAccessException) { throw; }
            catch (Exception) { throw; }
        }
        /// <summary>
        /// Setup the BackgroundWorker
        /// </summary>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        private void SetupTimer()
        {
            try
            {
                lock (this)
                {
                    this.LastFileModification = File.GetLastWriteTime(this.FilePath);
                    vTimer.Elapsed += new ElapsedEventHandler(Event_TimerElapsed);
                }
            }
            catch (ArgumentException) { throw; }
            catch (NotSupportedException) { throw; }
            catch (PathTooLongException) { throw; }
            catch (UnauthorizedAccessException) { throw; }
            catch (Exception) { throw; }
        }
        #endregion

        #region Events
        /// <summary>
        /// Thrown if the file changed. Check if the last modification time changed
        /// </summary>
        public event FileChangedEventHandler FileChanged;
        /// <summary>
        /// Notify a file changed
        /// </summary>
        /// <param name="args"></param>
        protected void OnFileChanged(FileChangedEventArgs args)
        {
            if (FileChanged != null)
                FileChanged(this, args);
        }
        /// <summary>
        /// Occures when the timer ticks
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Event_TimerElapsed(object sender, ElapsedEventArgs e)
        {
            lock (this)
            {
                DateTime dtBuffer = File.GetLastWriteTime(this.FilePath);
                if (dtBuffer > this.LastFileModification)
                {
                    this.LastFileModification = dtBuffer;
                    FileChangedEventArgs a = new FileChangedEventArgs();
                    a.DateTimeOfChange = dtBuffer;
                    a.DateTimeOfEvent = DateTime.Now;
                    OnFileChanged(a);
                }
            }
        }
        #endregion

        #region IEnumerator Members
        /// <summary>
        /// Return the current element
        /// </summary>
        public object Current { get { return vList[vIndex]; } }
        /// <summary>
        /// Move to the next element
        /// </summary>
        /// <returns></returns>
        public bool MoveNext() { return ++vIndex < vList.Count; }
        /// <summary>
        /// Reset the index of the enumaration
        /// </summary>
        public void Reset() { vIndex = -1; }
        /// <summary>
        /// Get the enumerator
        /// </summary>
        /// <returns></returns>
        public IEnumerator GetEnumerator() { return new LogSet(); }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Dispose the ressources
        /// </summary>
        public void Dispose()
        {
            this.vTimer.Stop();
            this.vTimer.Dispose();            
        }
        #endregion
    }
}
